#!/usr/bin/python
#
# Copyright (C) 2008-2011 W. Trevor King <wking@drexel.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

# Version 0.2, W. Trevor King, Feb. 20th, 2008
# (added OptionParser code and stripchart debug parameters)
# Version 0.3, W. Trevor King, Mar. 25th, 2010
# (added multiple channels option)
# Version 0.4, W. Trevor King, Nov. 12th, 2011
# Transition from optparse -> argparse and update to pycomedi 0.3.

"""Simple analog channel polling script.

Spends most of its time sleeping, waking up every `dt` seconds to
record the current voltage.
"""

import logging as _logging
import os.path as _os_path
import sys as _sys
import time as _time

try:
    import numpy as _numpy
    from matplotlib import pyplot as _pyplot
except (ImportError, RuntimeError), _matplotlib_import_error:
    _pyplot = None

import pycomedi.device as _pycomedi_device
import pycomedi.subdevice as _pycomedi_subdevice
import pycomedi.channel as _pycomedi_channel
import pycomedi.constant as _pycomedi_constant
try:
    from pypid.backend.melcor import MelcorBackend as _TemperatureBackend
except ImportError, pypid_import_error:
    _TemperatureBackend = None


__version__ = '0.4'


class Monitor (object):
    """Take measurements on a single channel every dt seconds.

    Save '<time>\t<bit_val>\t<physical_val>\n' records.
    """

    def __init__(self, channels=[0], temperature=False,
                 ambient_temperature=False, dt=4, comedi_device='/dev/comedi0',
                 data_stream=None, logger=None, plotter=None):
        self.channel_indexes = channels
        self.with_temperature = temperature
        self.with_ambient_temperature = ambient_temperature
        self.dt = dt
        self.comedi_device = comedi_device
        self.data_stream = data_stream
        self._logger = logger
        self._plotter = plotter
        self.temperature = None

    def log(self, level=_logging.DEBUG, msg=None):
        assert msg is not None
        if self._logger:
            self._logger.log(level, msg)

    def run(self):
        try:
            self._setup()
            self._run()
        finally:
            self._teardown()

    def _setup(self):
        self.log(msg='setup slow bend monitor')
        self.log(msg='chan {}, dt {}, data_stream {}'.format(
                self.channel_indexes, self.dt, self.data_stream))
        self._setup_channels()
        if self.with_temperature or self.with_ambient_temperature:
            self._setup_temperature()
        self._make_header()

    def _setup_channels(self):
        self.log(msg='setup input channels')
        self.device = _pycomedi_device.Device(self.comedi_device)
        self.device.open()
        self.subdevice = self.device.find_subdevice_by_type(
            _pycomedi_constant.SUBDEVICE_TYPE.ai,
            factory=_pycomedi_subdevice.StreamingSubdevice)
        self.channels = []
        for c_index in self.channel_indexes:
            c = self.subdevice.channel(
                c_index,
                factory=_pycomedi_channel.AnalogChannel,
                aref=_pycomedi_constant.AREF.diff)
            c.range = c.find_range(
                unit=_pycomedi_constant.UNIT.volt,
                min=-10, max=10)
            self.channels.append(c)
        self.converters = [c.get_converter() for c in self.channels]

    def _setup_temperature(self):
        self.log(msg='setup temperature channel')
        if _TemperatureBackend is None:
            raise pypid_import_error
        self.temperature = _TemperatureBackend()

    def _teardown(self):
        self.log(msg='teardown slow bend monitor')
        self._teardown_channels()
        self._teardown_temperature()

    def _teardown_channels(self):
        self.log(msg='teardown input channels')
        if hasattr(self, 'device') and self.device is not None:
            self.device.close()
            self.device = self.subdevice = self.channels = None

    def _teardown_temperature(self):
        self.log(msg='teardown temperature channel')
        if self.temperature is not None:
            self.temperature.cleanup()
            self.temperature = None

    def _run(self):
        self.log(msg='running slow bend monitor')
        self.start_time = _time.time()
        next_read_time = self.start_time
        self.log(_logging.INFO, 'press Control-c to stop acquisition')
        try:
            while True:
                tm,bitvals,physicals = self._take_and_save_reading()
                next_read_time += self.dt
                self._wait_until(next_read_time)
        except KeyboardInterrupt:
            pass

    def _take_and_save_reading(self):
        tm = _time.time()
        bitvals,physicals = self._measure_voltage()
        if self.with_temperature:
            bitval,physical = self._measure_temperature()
            bitvals.insert(0, bitval)
            physicals.insert(0, physical)
        if self.with_ambient_temperature:
            bitval,physical = self._measure_ambient_temperature()
            bitvals.insert(0, bitval)
            physicals.insert(0, physical)
        self._save_reading(tm, bitvals, physicals)
        return (tm, bitvals, physicals)

    def _measure_voltage(self):
        self.log(msg='measure voltage')
        bitvals = [c.data_read_delayed(nano_sec=1e3) for c in self.channels]
        self.log(msg='read bit values: {}'.format(
                '\t'.join('{:g}'.format(b) for b in bitvals)))
        physicals = [cv.to_physical(b)
                     for cv,b in zip(self.converters, bitvals)]
        return (bitvals, physicals)

    def _measure_temperature(self):
        self.log(msg='measure temperature')
        physical = self.temperature.get_pv()
        bitvalue = self.temperature._read('CHANNEL_1_A_D_COUNTS')
        return (bitvalue, physical)

    def _measure_ambient_temperature(self):
        self.log(msg='measure ambient temperature')
        physical = self.temperature.get_ambient_pv()
        bitvalue = self.temperature._read('AMBIENT_A_D_COUNTS')
        return (bitvalue, physical)

    def _make_header(self):
        'Create the save the data header'
        fields = ['time (second)']
        name_units = [('chan {}'.format(c.index), c.range.unit.name)
                      for c in self.channels]
        if self.with_temperature:
            name_units.insert(0, ('temperature', self.temperature.pv_units))
        if self.with_ambient_temperature:
            name_units.insert(0, ('ambient temperature', self.temperature.pv_units))
        for name,unit in name_units:
            fields.extend(['{} ({})'.format(name, u) for u in ['bit', unit]])
        headline = '#{}'.format('\t'.join(fields))
        self.data_stream.write(headline + '\n')
        self.data_stream.flush()
        self.log(_logging.INFO, headline)
        if self._plotter:
            self._plotter.add_header(fields)

    def _save_reading(self, time, bitvals, physicals):
        self.log(msg='save measurement')
        dataline = '{:g}\t{}'.format(
            time - self.start_time,
            '\t'.join('{:d}\t{:g}'.format(b, p)
                      for b,p in zip(bitvals, physicals)))
        self.data_stream.write(dataline + '\n')
        self.data_stream.flush()
        self.log(_logging.INFO, dataline)
        if self._plotter:
            self._plotter.add_points(time, bitvals, physicals)

    def _wait_until(self, target_time):
        self.log(msg='sleep until {}'.format(target_time))
        dt = target_time - _time.time()
        if dt > 0:
            _time.sleep(dt)


class Plotter (object):
    """Matplotlib-based strip-chart
    """
    def __init__(self, count=100):
        self.count = count
        if _pyplot is None:
            raise _matplotlib_import_error
        _pyplot.ion()

    def add_header(self, fields):
        self.figure = _pyplot.figure()
        self.channels = len(fields)/2  # integer division
        self.axes = []
        self.lines = []
        for i in range(self.channels):
            self.axes.append(self.figure.add_subplot(self.channels, 1, i+1))
            self.axes[i].set_title(fields[2*i+1])
            self.lines.append(self.axes[i].plot(
                    range(self.count),
                    _numpy.zeros((self.count,), dtype=_numpy.float), 'r.'))
        #_pyplot.show(block=False)  # block is an experimental kwarg
        _pyplot.show()

    def add_points(self, time, bit_values, physical_values):
        for i in range(self.channels):
            phys_values = self.lines[i][0].get_ydata()
            phys_values = _numpy.roll(phys_values, -1)
            phys_values[-1] = physical_values[i]
            self.lines[i][0].set_ydata(phys_values)
            self.axes[i].relim()
            self.axes[i].autoscale(axis='y')
        self.figure.canvas.draw()

def _get_data_stream(data_dir=None, logger=None):
    if data_dir is None:
        return _sys.stdout
    while True:
        timestamp = _time.strftime('%Y-%m-%d_%H-%M-%S')
        data_path = _os_path.join(args.data_dir, timestamp)
        if not _os_path.exists(data_path):
            break
        if logger:
            logger.warning(
                '{} already exists, wait a second and try again'.format(
                    data_path))
        _time.sleep(1)  # try the next second
    return open(data_path, 'w')

def _get_logger(level=_logging.DEBUG):
    logger = _logging.getLogger('slow_bend')
    logger.setLevel(level)
    ch = _logging.StreamHandler()
    ch.setLevel(level)
    formatter = _logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


if __name__ == '__main__':
    from argparse import ArgumentParser

    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        '--version', action='version', version=__version__)
    parser.add_argument(
        '-t', '--timestep', dest='dt', type=float, default=4.0,
        help='Measure voltage evert DT seconds (default: %(default)s)')
    parser.add_argument(
        dest='channels', nargs='+', metavar='CHANNEL', type=int,
        help='Input voltage CHANNEL(S)')
    parser.add_argument(
        '-d', '--data-directory', dest='data_dir', metavar='DATA_DIR',
        help='Write output to subdir of DATA_DIR')
    parser.add_argument(
        '-T', '--temperature', dest='temperature',
        default=False, action='store_const', const=True,
        help='Also record the temperature (thermocouple)')
    parser.add_argument(
        '-a', '--ambient-temperature', dest='ambient_temperature',
        default=False, action='store_const', const=True,
        help='Also record the ambient (room) temperature')
    parser.add_argument(
        '-p', '--plot', dest='plot',
        default=False, action='store_const', const=True,
        help='Display recorded physical values on a strip-chart')
    parser.add_argument(
        '-v', '--verbose', dest='verbose',
        default=0, action='count',
        help='Print debugging information (repeat to increase verbosity)')

    args = parser.parse_args()

    if args.verbose:
        if args.verbose > 1:
            level = _logging.DEBUG
        else:
            level = _logging.INFO
        logger = _get_logger(level=level)
    else:
        logger = None
    if args.plot:
        plotter = Plotter()
    else:
        plotter = None

    data_stream = _get_data_stream(data_dir=args.data_dir, logger=logger)
    try:
        m = Monitor(
            channels=args.channels, temperature=args.temperature,
            ambient_temperature=args.ambient_temperature, dt=args.dt,
            data_stream=data_stream, logger=logger, plotter=plotter)
        m.run()
    finally:
        if data_stream != _sys.stdout:
            data_stream.close()
